/*
  Alarm.m
  Copyright (c) 1996-1999 Dwight D. Everhart.  All rights reserved.

  This source code has been released to the public.  Use is governed
  by the MiscKit license, which basically means you can use it for
  whatever purpose you like, commercial or non-commercial.  This code
  is provided AS IS.  No warranty is expressed or implied.
*/

#import "Alarm.h"
#import "Date.h"
#import "setInstance.h"
#import <AppKit/AppKit.h>
#import <Foundation/Foundation.h>


#define CURRENT_VERSION 0


static NSMutableArray *alarms;
static NSMutableArray *freeList;
static BOOL executingAlarms = NO;
static BOOL freeing = NO;
static BOOL debugAlarms = NO;



static void freeClassData (void)
{
   [Alarm freeClassVariables];
}



@implementation Alarm



+ (void) initialize
{
   [super initialize];

   if ([self class] == [Alarm class]) {
      atexit (freeClassData);
      [self setVersion: CURRENT_VERSION];
   }
}



+ (BOOL) isScheduled: (Alarm *) alarm
{
   return alarm && alarms && [alarms containsObject: alarm];
}



+ (BOOL) scheduleAlarm: (Alarm *) alarm
{
   if (!alarm) return YES;
   if (freeing) return NO;

   if (!alarms) {
      if (debugAlarms) NSLog (@"Alarm: Creating alarm list");
      alarms = [[NSArray allocWithZone: [self zone]] initWithCapacity: 8];
   }

   if (![self isScheduled: alarm]) {
      if (debugAlarms) {
         NSLog (@"Alarm: Adding %p (%@) to alarm list",
                alarm, [[alarm alarmTime] description]);
      }

      [alarms addObject: alarm];
   } else {
      /* The time of the alarm may have changed. */
      if (debugAlarms) {
         NSLog (@"Alarm: Re-sorting alarm list for %p (%@)",
                alarm, [[alarm alarmTime] description]);
      }

      [alarms sortUsingSelector: @selector (compare:)];
   }

   /*
      Schedule an immediate execution if the alarm needs to be executed
    immediately.  Otherwise, schedule a regular execution if the alarm
    is the first one to be added to the list since it was last emptied.
    */
   if ([[alarm alarmTime] compare: [Date now]] <= 0) {
      if (debugAlarms) {
         NSLog (@"Alarm: Scheduling an immediate execution for late alarm %p (%@)",
                alarm, [[alarm alarmTime] description]);
      }

      [NSObject cancelPreviousPerformRequestsWithTarget: self
                                               selector: @selector (executeAlarmsImmediately:) object: self];
      [self performSelector: @selector (executeAlarmsImmediately:)
                 withObject: self afterDelay: 0];
   } else if ([alarms count] == 1) {
      [self scheduleNextExecution];
   }

   return YES;
}



+ (void) unscheduleAlarm: (Alarm *) alarm
{
   if (!freeing && [self isScheduled: alarm]) {
      if (debugAlarms) {
         NSLog (@"Alarm: Removing %p (%@) from alarm list",
                alarm, [[alarm alarmTime] description]);
      }

      [alarms removeObject: alarm];
      if ([alarms count] == 0) {
         [self cancelExecutions];
         [self freeClassVariables];
      }
   }
}



+ (BOOL) executeAlarms: (id) sender
{
   NSMutableArray *executionList;
   Date *now;
   Alarm *alarm;
   int i;

   if (debugAlarms) NSLog (@"Alarm: executeAlarms called");
   if ([alarms count] == 0) return YES;
   if (freeing) return NO;

   executionList = [NSMutableArray arrayWithCapacity: 8];

   if (executingAlarms) return YES;
   executingAlarms = YES;

   do {
      /* Get the current time. */
      now = [Date now];
      if (debugAlarms) {
         NSLog (@"Alarm: Current time is %@", [now description]);
      }

      /* Build the list of alarms that need to be executed. */
      [executionList removeAllObjects];
      for (i = 0; i < (const) [alarms count]; i++) {
         alarm = [alarms objectAtIndex: i];
         if ([[alarm alarmTime] compare: now] <= 0) break;
         if (debugAlarms) {
            NSLog (@"Alarm: Adding %p (%@) to execution list",
                   alarm, [[alarm alarmTime] description]);
         }

         [executionList addObject: alarm];
      }

      for (i = 0; (alarm = [executionList objectAtIndex: i]); i++) {
         NSParameterAssert ([alarms indexOfObject: alarm] == 0);
         if (debugAlarms) NSLog (@"Alarm: Executing %p", alarm);
         [alarm executeAlarm];

         /*
            We need this check in case the alarm was rescheduled
          as part of its execution.
          */
         if ([[alarm alarmTime] compare: now] <= 0) {
            if (debugAlarms) {
               NSLog (@"Alarm: Removing %p from alarms list", alarm);
            }

            [alarms removeObject: alarm];
         } else if (debugAlarms) {
            if ([alarms containsObject: alarm]) {
               NSLog (@"Alarm: %p was rescheduled", alarm);
            } else {
               NSLog (@"Alarm: %p was removed", alarm);
            }
         }
      }
   } while ([executionList count] > 0);

   if ([alarms count] > 0) {
      [self scheduleNextExecution];
   } else {
      [self freeClassVariables];
   }

   executingAlarms = NO;
#warning This freeList busness is crud left over from NEXTSTEP and needs to be fixed.
   //   [freeList freeObjects];
   return YES;
}



+ (BOOL) executeAlarmsImmediately: (id) sender
{
   return [self executeAlarms: sender];
}



+ (void) scheduleNextExecution
{
   int seconds = 60 - [[NSCalendarDate calendarDate] secondOfMinute];

   if (freeing) return;

   [NSObject cancelPreviousPerformRequestsWithTarget: self
                                            selector: @selector (executeAlarms:) object: self];
   [self performSelector: @selector (executeAlarms:)
              withObject: self afterDelay: seconds];

   if (debugAlarms) {
      NSLog (@"Alarm: Scheduled next execution for %i seconds from now",
             seconds);
   }
}



+ (void) cancelExecutions
{
   if (debugAlarms) NSLog (@"Alarm: Canceling execution request");
   [NSObject cancelPreviousPerformRequestsWithTarget: self
                                            selector: @selector (executeAlarms:) object: self];
}



+ (void) freeAlarmAfterExecution: (Alarm *) freeMe
{
   if (freeing) return;

   if (!freeList) {
      if (debugAlarms) NSLog (@"Alarm: Creating free list");
      freeList = [[NSMutableArray allocWithZone: [self zone]] initWithCapacity: 8];
   }

   [freeList addObject: freeMe];
}



+ (void) freeClassVariables
{
   if (freeing) return;
   if (debugAlarms) NSLog (@"Alarm: Freeing class variables");
   freeing = YES;
   [alarms release];
#warning This freeList busness is crud left over from NEXTSTEP and needs to be fixed.
   //   [freeList freeObjects];
   [freeList release];
   alarms = nil;
   freeList = nil;
}



- (id) init
{
   return [self initWithAlarmTime: nil target: nil andMessage: NULL];
}



- (id) initWithAlarmTime: (Date *) theAlarmTime
                  target: target andMessage: (SEL) message
{
   if (!(self = [super init])) return nil;
   [self setAlarmTime: theAlarmTime target: target andMessage: message];
   return self;
}



- (id) initWithCoder: (NSCoder *) aDecoder
{
   self = [super init];
   if (!self) return nil;

   alarmTime = [[aDecoder decodeObject] retain];
   alarmTarget = [[aDecoder decodeObject] retain];
   [aDecoder decodeValueOfObjCType: @encode (SEL) at: &alarmMessage];

   if ([self isScheduleable]) {
      [Alarm scheduleAlarm: self];
   }

   return self;
}



- (void) encodeWithCoder: (NSCoder *) aCoder
{
   [aCoder encodeObject:alarmTime];
   [aCoder encodeConditionalObject:alarmTarget];
   [aCoder encodeValueOfObjCType: @encode (SEL) at: &alarmMessage];
}



- (id) copyWithZone: (NSZone *) zone
{
   /*
   Don't schedule the alarm, since that would only cause two identical
    messages to be sent to the same object, one right after the other.
    */

   return [[Alarm allocWithZone: zone]
           initWithAlarmTime: [alarmTime copyWithZone: zone]
                      target: alarmTarget
                  andMessage: alarmMessage];
}



- (void) dealloc
{
   if (executingAlarms && executing) {
      [Alarm freeAlarmAfterExecution: self];
   } else {
      [Alarm unscheduleAlarm: self];
      [alarmTime release];
      [super dealloc];
   }
}



- (BOOL) isScheduled
{
   return [Alarm isScheduled: self];
}



- (BOOL) isScheduleable
{
   return alarmTime && alarmTarget && alarmMessage;
}



- (BOOL) isExecuting
{
   return executing;
}



- (Date *) alarmTime
{
   return alarmTime;
}



- (id) alarmTarget
{
   return alarmTarget;
}



- (SEL) alarmMessage
{
   return alarmMessage;
}



- (void) setAlarmTime: (Date *) newTime
{
   [self setAlarmTime: newTime target: alarmTarget andMessage: alarmMessage];
}



- (void) setAlarmTarget: (id) newTarget
{
   [self setAlarmTime: alarmTime target: newTarget andMessage: alarmMessage];
}



- (void) setAlarmMessage: (SEL) newMessage
{
   [self setAlarmTime: alarmTime target: alarmTarget andMessage: newMessage];
}



- (void) setAlarmTime: (Date *) newTime
               target: (id) newTarget andMessage: (SEL) newMessage
{
   if (newTime) {
      if (!alarmTime) {
         alarmTime = [[Date allocWithZone: [self zone]] init];
      }

      [alarmTime setDate: newTime];
   } else {
      [alarmTime release];
      alarmTime = nil;
   }

   setInstance (alarmTarget, newTarget);
   alarmMessage = newMessage;

   if ([self isScheduleable]) {
      [Alarm scheduleAlarm: self];
   } else {
      [Alarm unscheduleAlarm: self];
   }
}



- (void) executeAlarm
{
   if ([alarmTarget respondsToSelector: alarmMessage]) {
      executing = YES;
      [alarmTarget performSelector: alarmMessage withObject: self];
      executing = NO;
   } else {
      NSLog (@"Alarm: Target (%@ %p) does not respond to message (%@)",
             [[alarmTarget class] name], alarmTarget,
             NSStringFromSelector (alarmMessage));
   }

   /* Don't remove the alarm from the list here.  Let +executeAlarms: do it. */
}



- (NSComparisonResult) compare: (id) anObject
{
   Date *anObjectTime;
   NSComparisonResult result;

   NSParameterAssert ([anObject isKindOfClass: [Alarm class]]);

   if (anObject == self) return NSOrderedSame;

   anObjectTime = [anObject alarmTime];
   if (alarmTime && anObjectTime) {
      result = [alarmTime compare: [(Alarm *) anObject alarmTime]];
   } else if (!alarmTime && !anObjectTime) {
      result = NSOrderedSame;
   } else if (alarmTime && !anObjectTime) {
      result = NSOrderedAscending;
   } else {
      result = NSOrderedDescending;
   }

   // This next bit is pretty silly, but we need to have some way to distinguish
   // between instances which are the same except for their target or message.
   if (result == NSOrderedSame) {
      id anObjectTarget = [anObject alarmTarget];
      if (alarmTarget == anObjectTarget) {
         SEL anObjectMessage = [anObject alarmMessage];
         if (alarmMessage == anObjectMessage) {
            result = NSOrderedSame;
         } else if (alarmMessage < anObjectMessage) {
            result = NSOrderedAscending;
         } else {
            result = NSOrderedDescending;
         }
      } else if (alarmTarget < anObjectTarget) {
         result = NSOrderedAscending;
      } else {
         result = NSOrderedDescending;
      }
   }

   return result;
}



- (BOOL) isEqual: (id) anObject
{
   return [self compare: anObject] == NSOrderedSame;
}



@end
