/*
  Appointment.m
  Copyright (c) 1996-1999 Dwight D. Everhart.  All rights reserved.

  This source code has been released to the public.  Use is governed
  by the MiscKit license, which basically means you can use it for
  whatever purpose you like, commercial or non-commercial.  This code
  is provided AS IS.  No warranty is expressed or implied.
*/

#import "Appointment.h"
#import "Date.h"
#import "setInstance.h"
#import <Foundation/NSAttributedString.h>
#import <Foundation/NSException.h>
#import <Foundation/NSString.h>


#define REASONABLE_DURATION (24 * 60)


static int weekStart = 0;



@implementation Appointment



- (id) init
{
   self = [super init];
   if (!self) return nil;

   startDate = [[Date allocWithZone:[self zone]] init];
   endDate   = [[Date allocWithZone:[self zone]] init];

   interval  = RI_NONE;
   frequency = 1;
   duration = 60;
   scheduleLevel = ASL_EXACT;

   return self;
}



- (id) copyWithZone: (NSZone *) zone
{
   Appointment *clone = NSCopyObject (self, 0, zone);

   if (!clone) return nil;
   clone->title           = [title copyWithZone:zone];
   clone->descriptionText = [descriptionText copyWithZone: zone];
   clone->startDate       = [startDate copyWithZone: zone];
   clone->endDate         = [endDate   copyWithZone: zone];
   clone->interval        = RI_NONE;
   clone->frequency       = 1;
   return clone;
}



- (void) dealloc
{
   [title release];
   [descriptionText release];
   [startDate release];
   [endDate release];
   [super dealloc];
}



@end



@implementation Appointment  (AppointmentAttributes)



- (NSString *) title
{
   return title;
}



- (NSAttributedString *) descriptionText
{
   return descriptionText;
}



- (Date *) startDate
{
   return startDate;
}


- (Date *) endDate
{
   return endDate;
}



- (int) interval
{
   return interval;
}



- (int) frequency
{
   return frequency;
}



- (int) duration
{
   return duration;
}



- (int) scheduleLevel
{
   return scheduleLevel;
}



- (BOOL) isRegular
{
   return ((scheduleLevel == ASL_EXACT) &&
           ((interval == RI_NONE) || (frequency == 1)));
}



- (BOOL) getNextOccuranceDate: (Date *) occuranceDate
{
   NSParameterAssert (occuranceDate);

   if ([occuranceDate minuteOfDay] > [startDate minuteOfDay]) {
      [occuranceDate incrementDay];
   }

   if (scheduleLevel == ASL_EXACT) {
      [occuranceDate setMinute: [startDate minuteOfDay]];
   } else {
      [occuranceDate setMinute: 0];
   }

   if ([occuranceDate compare: endDate] > 0) {
      /*
       This branch must come first; otherwise, nil won't be returned
       if the occuranceDate is later than the endDate.
       */
      [occuranceDate setDate: endDate];
      return NO; /* To indicate that there is no next occurance. */
   } else if ((interval == RI_NONE) || (frequency == 0)) {
      /* Extra check here in case endDate is out of sync with startDate. */
      if ([occuranceDate compare: startDate] > 0) return NO;
      [occuranceDate setDate: startDate];
   } else if ([occuranceDate compare: startDate] < 0) {
      [occuranceDate setDate: startDate];
   } else {
      while (![self isScheduledFor: (Date *) occuranceDate] &&
             ([occuranceDate compare: endDate] < 0)) {
         [occuranceDate incrementDay];
      }
   }

   return YES;
}



- (BOOL) isScheduledFor: (Date *) date
{
   if (!date || ([date compare: startDate] < 0) ||
       ([date compare: endDate] > 0)) {
      return NO;
   }

   switch (interval) {
      case RI_NONE:
         return ([startDate compare: date] == 0);
      case RI_DAILY:
         return ((frequency == 1) ||
                 ([startDate daysUntil: date] % frequency) == 0);
      case RI_WEEKLY:
         return (([startDate weekday] == [date weekday]) &&
                 ((frequency == 1) ||
                  (([startDate weeksUntil: date] % frequency) == 0)));
      case RI_MONTHLY:
         return (([startDate dayOfMonth] == [date dayOfMonth]) &&
                 ((frequency == 1) ||
                  (([startDate monthsUntil: date] % frequency) == 0)));
      case RI_YEARLY:
         return ((([startDate dayOfMonth] == [date dayOfMonth]) &&
                  ([startDate monthOfYear] == [date monthOfYear])) &&
                 ((frequency == 1) ||
                  (([startDate yearsUntil: date] % frequency) == 0)));
      default:
         [NSException raise: NSInvalidArgumentException
                     format: @"Invalid repeat interval: %d", interval];
   }

   return NO;
}



- (NSComparisonResult) compare: (id) anObject
{
   Appointment *otherAppt;

   NSParameterAssert ([anObject isKindOfClass: [Appointment class]]);
   otherAppt = anObject;

   return [startDate compare: [otherAppt startDate]];
}



@end



@implementation Appointment (AppointmentActions)



- (void) setTitle: (NSString *) newTitle
{
   setInstanceByCopy (title, newTitle);
}



- (void) setDescriptionText: (NSAttributedString *) newDescriptionText
{
   setInstance (descriptionText, newDescriptionText);
}



- (void) setStartDate: (Date *) newStartDate andConstrain: (BOOL) constrain
{
   if ([startDate isEqual: newStartDate]) return;

   if ([endDate compare: newStartDate] < 0) {
      [endDate setDate: newStartDate];
   }

   if (constrain && ([startDate compare: newStartDate] < 0)) {
      switch (interval) {
         case RI_NONE:
            [startDate setDate: newStartDate];
            break;
         case RI_DAILY:
            if (frequency > 1) {
               [startDate changeDayBy: [startDate daysUntil: newStartDate] % frequency];
            }
            break;
         case RI_WEEKLY:
            [startDate setDayToNearestWeekday: [newStartDate weekday]];
            break;
         case RI_MONTHLY:
            [startDate setDay: [newStartDate dayOfMonth]];
            break;
         case RI_YEARLY:
            [startDate setMonth: [newStartDate monthOfYear]];
            [startDate setDay:   [newStartDate dayOfMonth]];
            break;
         default:
            [NSException raise: NSInvalidArgumentException
                        format: @"Invalid repeat interval: %d", interval];
            break;
      }

      if (scheduleLevel == ASL_EXACT) {
         [startDate setMinute: [newStartDate minuteOfDay]];
      } else {
         [startDate setMinute: 0];
      }
   } else {
      [startDate setDate: newStartDate];
   }

   [self constrainScheduleDateToScheduleLevel];
}



- (void) setEndDate: (Date *) newEndDate
{
   if (interval == RI_NONE) return;

   if ([endDate isEqual: newEndDate]) return;

   if ([startDate compare: newEndDate] > 0) {
      if (![endDate isEqual: startDate]) {
         [endDate setDate: startDate];
      }
   } else {
      [endDate setDate: newEndDate];
   }
}



- (void) setInterval: (int) newInterval
{
   if (interval != newInterval) {
      interval = newInterval;
      [self constrainScheduleDateToScheduleLevel];
   }
}



- (void) setFrequency: (int) newFrequency
{
   if (frequency != newFrequency) {
      frequency = newFrequency;
      [self constrainScheduleDateToScheduleLevel];
   }
}



- (void) setDuration: (int) newDuration
{
   if (duration != newDuration) {
      if (duration > REASONABLE_DURATION) {
         [NSException raise: NSInvalidArgumentException
                     format: @"Attempt to set appointment duration to %hd", duration];
      }

      duration = newDuration;
   }
}



- (void) setScheduleLevel: (int) newScheduleLevel
{
   if (scheduleLevel != newScheduleLevel) {
      scheduleLevel = newScheduleLevel;
      [self constrainScheduleDateToScheduleLevel];
   }
}



- (void) constrainScheduleDateToScheduleLevel
{
   switch (scheduleLevel) {
      case ASL_YEAR:
         [startDate setMonth:  1];
         [startDate setDay:    1];
         [startDate setMinute: 0];
         if ((interval != RI_NONE) && (interval < RI_YEARLY)) {
            interval = RI_YEARLY;
            frequency = 1;
         }
            break;
      case ASL_MONTH:
         [startDate setDay:    1];
         [startDate setMinute: 0];
         if ((interval != RI_NONE) && (interval < RI_MONTHLY)) {
            interval = RI_MONTHLY;
            frequency = 1;
         }
            break;
      case ASL_WEEK:
         [startDate setDayToThisOrPreviousWeekday: weekStart];
         [startDate setMinute: 0];
         if ((interval != RI_NONE) && (interval < RI_WEEKLY)) {
            interval = RI_WEEKLY;
            frequency = 1;
         }
            break;
      case ASL_DAY:
         [startDate setMinute: 0];
         break;
   }

   // Keep the endDate the same as the startDate for non-repeating appointemnts.
   if ((interval == RI_NONE) || (frequency == 0) ||
       ([startDate compare: endDate] > 0)) {
      [endDate setDate: startDate];
   }
}



- (void) bumpWithSelector: (SEL) bumpSelector
{
   [startDate performSelector: bumpSelector];
   [endDate   performSelector: bumpSelector];
}



- (void) bumpToNextDay
{
   [self bumpWithSelector: @selector (incrementDay)];
}



- (void) bumpToNextWeek
{
   [self bumpWithSelector: @selector (incrementWeek)];
}



- (void) bumpToNextMonth
{
   [self bumpWithSelector: @selector (incrementMonth)];
}



- (void) bumpToNextYear
{
   [self bumpWithSelector: @selector (incrementYear)];
}



@end
