/*
  Date+Actions.m
  Copyright (c) 1996-1999 Dwight D. Everhart.  All rights reserved.

  This source code has been released to the public.  Use is governed
  by the MiscKit license, which basically means you can use it for
  whatever purpose you like, commercial or non-commercial.  This code
  is provided AS IS.  No warranty is expressed or implied.
*/

#import "Date.h"
#import <Foundation/NSDate.h>
#import <Foundation/NSException.h>
#import <Foundation/NSString.h>

#ifdef GNUSTEP
#import <Foundation/NSCalendarDate.h>
#endif

typedef enum {DC_DAY, DC_MONTH, DC_YEAR} dateComponent_t;



@implementation Date (DateActions)



- (void) setMinute: (int) newMinute
{
   NSParameterAssert (newMinute <= MINUTES_PER_DAY);
   minute = newMinute;
}



- (void) setDay: (int) aDay
{
   [[self class] validateYear: year month: month day: aDay];
   day = aDay;
}



- (void) setMonth: (int) newMonth
{
   int daysInMonth = [[self class] numberOfDaysInMonth: newMonth forYear: year];
   int newDay = (day > daysInMonth) ? daysInMonth : day;

   [self setYear: year month: newMonth day: newDay];
}



- (void) setYear: (int) aYear
{
   [[self class] validateYear: aYear month: month day: day];
   year = aYear;
}



- (void) setYear: (int) aYear month: (int) aMonth day: (int) aDay
{
   [[self class] validateYear: aYear month: aMonth day: aDay];
   year = aYear;
   month = aMonth;
   day = aDay;
}



- (void) setYear: (int) aYear month: (int) aMonth day: (int) aDay minute: (int) aMinute
{
   [self setYear: aYear month: aMonth day: aDay];
   minute = aMinute;
}



- (void) setDate: (Date *) otherDate
{
   [self setYear: [otherDate yearOfCommonEra]
           month: [otherDate monthOfYear]
             day: [otherDate dayOfMonth]
          minute: [otherDate minuteOfDay]];
}



- (void) setDateToTM: (const struct tm *) tm
{
   NSParameterAssert (tm);
   [self setYear: tm->tm_year + 1900
           month: tm->tm_mon + 1
             day: tm->tm_mday
          minute: (tm->tm_hour * 60) + tm->tm_min];
}



- (void) setDateToTime_t: (time_t) newTime
{
   [self setDateToTM: localtime (&newTime)];
}



- (void) setDateToTimeIntervalSince1970: (NSTimeInterval) newInterval
{
   return [self setDateToTime_t: newInterval];
}



- (void) setDateToNSDate: (NSDate *) nsDate
{
   return [self setDateToTimeIntervalSince1970: [nsDate timeIntervalSince1970]];
}



- (void) setDateToCalendarDate: (NSCalendarDate *) calendarDate
{
  return [self setYear: [calendarDate yearOfCommonEra]
		month: [calendarDate monthOfYear]
		day: [calendarDate dayOfMonth]
		minute: ([calendarDate hourOfDay] * 60) + [calendarDate minuteOfHour]];
}



- (void) setDateToToday
{
   time_t now;

   time (&now);
   [self setDateToTM: localtime (&now)];
}



- (void) setToEarliestDate
{
   [self setYear: FIRST_YEAR month: FIRST_MONTH day: 1 minute: 0];
}



- (void) setToLatestDate
{
   [self setYear: LAST_YEAR month: LAST_MONTH day: 31 minute: MINUTES_PER_DAY - 1];
}



- (void) setDayOfYear:  (int) newDayOfYear
{
   [self changeDayBy: newDayOfYear - [self dayOfYear]];
}



- (void) setWeekOfYear: (int) newWeekOfYear
{
   [self changeWeekBy: newWeekOfYear - [self weekOfYear]];
}



- (void) setDayToPreviousDay: (int) newDay
{
   if (newDay >= day) {
      do {
         [self decrementMonth];
      } while (newDay > [self numberOfDaysInMonth]);
   }

   [self setDay: newDay];
}



- (void) setDayToNextDay: (int) newDay
{
   if (newDay <= day) {
      do {
         [self incrementMonth];
      } while (newDay > [self numberOfDaysInMonth]);
   }

   [self setDay: newDay];
}



- (void) setDayToThisOrPreviousDay: (int) newDay
{
   if (newDay > day) {
      do {
         [self decrementMonth];
      } while (newDay > [self numberOfDaysInMonth]);
   }

   [self setDay: newDay];
}



- (void) setDayToThisOrNextDay: (int) newDay
{
   if (newDay < day) {
      do {
         [self incrementMonth];
      } while (newDay > [self numberOfDaysInMonth]);
   }

   [self setDay: newDay];
}



- (void) setDayToPreviousWeekday: (int) weekday
{
   int delta = weekday - [self weekday];

   if (delta >= 0) delta -= DAYS_IN_WEEK;
   [self changeDayBy: delta];
}



- (void) setDayToNextWeekday: (int) weekday
{
   int delta = weekday - [self weekday];

   if (delta <= 0) delta += DAYS_IN_WEEK;
   [self changeDayBy: delta];
}



- (void) setDayToThisOrPreviousWeekday: (int) weekday
{
   int delta = weekday - [self weekday];

   if (delta > 0) delta -= DAYS_IN_WEEK;
   [self changeDayBy: delta];
}



- (void) setDayToThisOrNextWeekday: (int) weekday
{
   int delta = weekday - [self weekday];

   if (delta < 0) delta += DAYS_IN_WEEK;
   [self changeDayBy: delta];
}



- (void) setDayToNearestWeekday: (int) weekday
{
   int delta = weekday - [self weekday];

   if (delta > 3) {
      delta -= DAYS_IN_WEEK;
   } else if (delta < -3) {
      delta += DAYS_IN_WEEK;
   }

   [self changeDayBy: delta];
}



- (void) setMonthToPreviousMonth: (int) newMonth
{
   if (newMonth >= month) [self decrementYear];
   [self setMonth: newMonth];
}



- (void) setMonthToNextMonth: (int) newMonth
{
   if (newMonth <= month) [self incrementYear];
   [self setMonth: newMonth];
}



- (void) setMonthToThisOrPreviousMonth: (int) newMonth
{
   if (newMonth > month) [self decrementYear];
   [self setMonth: newMonth];
}



- (void) setMonthToThisOrNextMonth: (int) newMonth
{
   if (newMonth < month) [self incrementYear];
   [self setMonth: newMonth];
}



- (void) setDayToPreviousMonth: (int) newMonth andDay: (int) newDay
{
   if ((month < newMonth) || ((month == newMonth) && (day <= newDay))) {
      [self decrementYear];
   }

   while (![[self class] isValidYear: year month: newMonth day: newDay]) {
      [self decrementYear];
   }
}



- (void) setDayToNextMonth: (int) newMonth andDay: (int) newDay
{
   if ((month > newMonth) || ((month == newMonth) && (day >= newDay))) {
      [self incrementYear];
   }

   while (![[self class] isValidYear: year month: newMonth day: newDay]) {
      [self incrementYear];
   }
}



- (void) setDayToThisOrPreviousMonth: (int) newMonth andDay: (int) newDay
{
   if ((month < newMonth) || ((month == newMonth) && (day < newDay))) {
      [self decrementYear];
   }

   while (![[self class] isValidYear: year month: newMonth day: newDay]) {
      [self decrementYear];
   }
}



- (void) setDayToThisOrNextMonth: (int) newMonth andDay: (int) newDay
{
   if ((month > newMonth) || ((month == newMonth) && (day > newDay))) {
      [self incrementYear];
   }

   while (![[self class] isValidYear: year month: newMonth day: newDay]) {
      [self incrementYear];
   }
}



- (void) _changeComponent: (dateComponent_t) component by: (int) delta
{
   while (delta) {
      if (delta > 0) {
         switch (component) {
            case DC_DAY:
               [self incrementDay];
               break;
            case DC_MONTH:
               [self incrementMonth];
               break;
            case DC_YEAR:
               [self incrementYear];
               break;
            default:
               [NSException raise: NSInvalidArgumentException
                           format: @"Bad date component: %d", component];
               break;
         }
         delta--;
      } else {
         switch (component) {
            case DC_DAY:
               [self decrementDay];
               break;
            case DC_MONTH:
               [self decrementMonth];
               break;
            case DC_YEAR:
               [self decrementYear];
               break;
            default:
               [NSException raise: NSInvalidArgumentException
                           format: @"Bad date component: %d", component];
               break;
         }
         delta++;
      }
   }
}



- (void) changeMinuteBy: (int) delta
{
   int minuteChange = delta % MINUTES_PER_DAY;
   int dayChange = delta / MINUTES_PER_DAY;
   int newMinute = minute + minuteChange;

   if (newMinute < 0) {
      newMinute += MINUTES_PER_DAY;
      dayChange--;
   } else if (newMinute >= MINUTES_PER_DAY) {
      newMinute -= MINUTES_PER_DAY;
      dayChange++;
   }

   minute = newMinute;
   if (dayChange != 0) [self changeDayBy: dayChange];
}



- (void) incrementDay
{
   if (day + 1 > [self numberOfDaysInMonth]) {
      day = 1;
      [self incrementMonth];
   } else {
      day++;
   }
}



- (void) decrementDay
{
   if (day == 1) {
      [self decrementMonth];
      day = [self numberOfDaysInMonth];
   } else {
      day--;
   }
}



- (void) changeDayBy: (int) delta
{
   [self _changeComponent: DC_DAY by: delta];
}



- (void) incrementWeek
{
   [self changeDayBy: DAYS_IN_WEEK];
}



- (void) decrementWeek
{
   [self changeDayBy: -DAYS_IN_WEEK];
}



- (void) changeWeekBy: (int) delta
{
   [self changeDayBy: delta * DAYS_IN_WEEK];
}



- (void) incrementMonth
{
   int oldMonth = month;
   int oldDay = day;
   int days;

   if (month == LAST_MONTH) {
      month = FIRST_MONTH;
   } else {
      month++;
   }

   days = [[self class] numberOfDaysInMonth: month forYear: year];
   if (day > days) day = days;

   if (month == FIRST_MONTH) {
      [self incrementYear];
      month = oldMonth;
      day = oldDay;
   }
}



- (void) decrementMonth
{
   int oldMonth = month;
   int oldDay = day;
   int days;

   if (month == FIRST_MONTH) {
      month = LAST_MONTH;
   } else {
      month--;
   }

   days = [[self class] numberOfDaysInMonth: month forYear: year];
   if (day > days) day = days;

   if (month == LAST_MONTH) {
      [self decrementYear];
      month = oldMonth;
      day = oldDay;
   }
}



- (void) changeMonthBy: (int) delta
{
   [self _changeComponent: DC_MONTH by: delta];
}



- (void) incrementYear
{
   if (year == LAST_YEAR) return;

   if ([[self class]  isLeapYear: year] &&
       ![[self class] isLeapYear: year + 1] &&
       (month == FEBRUARY) && (day == 29)) {
      day = 28;
   }

   year++;
}



- (void) decrementYear
{
   if (year == FIRST_YEAR) return;

   if ([[self class]  isLeapYear: year] &&
       ![[self class] isLeapYear: year - 1] &&
       (month == FEBRUARY) && (day == 29)) {
      day = 28;
   }

   year--;
}



- (void) changeYearBy: (int) delta
{
   [self _changeComponent: DC_YEAR by: delta];
}



@end
