/*
  Date+Attributes.m
  Copyright (c) 1996-1999 Dwight D. Everhart.  All rights reserved.

  This source code has been released to the public.  Use is governed
  by the MiscKit license, which basically means you can use it for
  whatever purpose you like, commercial or non-commercial.  This code
  is provided AS IS.  No warranty is expressed or implied.
*/

#import "Date.h"
#import <Foundation/NSException.h>
#import <Foundation/NSString.h>

#ifdef GNUSTEP
#import <Foundation/NSCalendarDate.h>
#import <Foundation/NSTimeZone.h>
#endif



@implementation Date (DateAttributes)



- (int) minuteOfDay
{
   return minute;
}



- (int) dayOfMonth
{
   return day;
}



- (int) monthOfYear
{
   return month;
}



- (int) yearOfCommonEra
{
   return year;
}



- (int) weekday
{
   return [[self class] weekdayForYear: year month: month day: day];
}



- (int) dayOfYear
{
   int dayNumber = 0;
   int m;

   for (m = 1; m < month; m++) {
      dayNumber += [[self class] numberOfDaysInMonth: m forYear: year];
   }

   if ([self isLeapYear] && (month > FEBRUARY)) dayNumber++;
   dayNumber += day;
   return dayNumber;
}



- (int) weekOfYear
{
   int newYearsWeekday = [[self class] weekdayForYear: year month: 1 day: 1];
   int dayOfYear = [self dayOfYear];

   if (newYearsWeekday + dayOfYear - 1 < [[self class] weekStart]) {
      return WEEKS_IN_YEAR;
   } else {
      dayOfYear += newYearsWeekday - [[self class] weekStart] - 1;
      return ((dayOfYear / DAYS_IN_WEEK) % WEEKS_IN_YEAR) + 1;
   }
}



- (int) numberOfMonthsInYear
{
   return MONTHS_IN_YEAR;
}



- (int) numberOfWeeksInYear
{
   return WEEKS_IN_YEAR;
}



- (int) numberOfDaysInWeek
{
   return DAYS_IN_WEEK;
}



- (int) numberOfDaysInMonth
{
   return [[self class] numberOfDaysInMonth: month forYear: year];
}



- (int) numberOfDaysInYear
{
   return [[self class] numberOfDaysInYear: year];
}



- (int) earliestYear
{
   return [[self class] earliestYear];
}



- (int) latestYear
{
   return [[self class] latestYear];
}



- (int) startDayOfMonth
{
   return [[self class] weekdayForYear: year month: month day: 1];
}



- (BOOL) isLeapYear
{
   return [[self class] isLeapYear: year];
}



- (NSCalendarDate *) calendarDate
{
   return [NSCalendarDate dateWithYear: year month: month day: day
                                  hour: minute / 60 minute: minute % 60 second: 0
                              timeZone: [NSTimeZone localTimeZone]];
}



- (NSString *) description
{
   return [[self calendarDate] description];
}



- (int) minutesUntil: (Date *) otherDate
{
   Date *earlierDate;
   Date *laterDate;
   int minutes;

   if ([self compare: otherDate] < 0 ) {
      earlierDate = self;
      laterDate = otherDate;
   } else {
      earlierDate = otherDate;
      laterDate = self;
   }

   minutes = ([earlierDate daysUntil: laterDate] * MINUTES_PER_DAY) +
      ([laterDate minuteOfDay] - [earlierDate minuteOfDay]);

   if (laterDate == self) minutes = -minutes;
   return minutes;
}



- (int) minutesSince: (Date *) otherDate
{
   return -[self minutesUntil: otherDate];
}



- (int) daysUntil: (Date *) otherDate
{
   Date *earlierDate;
   Date *laterDate;
   int y, laterYear;
   int days;

   if ([self compare: otherDate] < 0 ) {
      earlierDate = self;
      laterDate = otherDate;
   } else {
      earlierDate = otherDate;
      laterDate = self;
   }

   laterYear = [laterDate yearOfCommonEra];
   for (days = 0, y = [earlierDate yearOfCommonEra]; y < laterYear; y++) {
      days += [[self class] numberOfDaysInYear: y];
   }

   days -= [earlierDate dayOfYear];
   days += [laterDate   dayOfYear];
   if (laterDate == self) days = -days;
   return days;
}



- (int) daysSince: (Date *) otherDate
{
   return -[self daysUntil: otherDate];
}



- (int) weeksUntil: (Date *) otherDate
{
   return [self daysUntil: otherDate] / DAYS_IN_WEEK;
}



- (int) weeksSince: (Date *) otherDate
{
   return -[self weeksUntil: otherDate];
}



- (int) monthsUntil: (Date *) otherDate
{
   return [self monthsWithDay: 1 until: otherDate];
}



- (int) monthsSince: (Date *) otherDate
{
   return [self monthsWithDay: 1 since: otherDate];
}



- (int) monthsWithDay: (int) aDay until: (Date *) otherDate
{
   Date *earlierDate;
   Date *laterDate;
   int y, m;
   int months;
   int earlierMonth, laterMonth;
   int earlierYear, laterYear;

   if ([self compare: otherDate] < 0 ) {
      earlierDate = self;
      laterDate = otherDate;
   } else {
      earlierDate = otherDate;
      laterDate = self;
   }

   earlierYear  = [earlierDate yearOfCommonEra];
   earlierMonth = [earlierDate monthOfYear];
   laterYear    = [laterDate yearOfCommonEra];
   laterMonth   = [laterDate monthOfYear];

   for (months = 0, y = earlierYear; y < laterYear; y++) {
      if ((aDay < 29) || ((aDay == 29) && [[self class] isLeapYear: y])) {
         months += 12;
      } else if (aDay < 31) {
         months += 11;
      } else {
         months += 7;
      }
   }

   if (aDay < 29) {
      months -= earlierMonth;
      months += laterMonth;
   } else {
      for (m = FIRST_MONTH; m <= earlierMonth; m++) {
         if ([[self class] isValidYear: earlierYear month: m day: aDay]) {
            months--;
         }
      }

      for (m = FIRST_MONTH; m <= laterMonth; m++) {
         if ([[self class] isValidYear: laterYear month: m day: aDay]) {
            months++;
         }
      }
   }

   if (laterDate == self) months = -months;
   return months;
}




- (int) monthsWithDay: (int) aDay since: (Date *) otherDate
{
   return -[self monthsWithDay: aDay until: otherDate];
}



- (int) yearsUntil: (Date *) otherDate
{
   return [otherDate yearOfCommonEra] - year;
}



- (int) yearsSince: (Date *) otherDate
{
   return year - [otherDate yearOfCommonEra];
}



- (unsigned int) hash
{
   return (((((year - FIRST_YEAR) * 1000) + [self dayOfYear]) * 10000) + minute);
}



- (BOOL) isEqual: (id) anObject
{
   Date *otherDate;

   if (anObject == self) return YES;
   if (![anObject isKindOfClass: [Date class]]) return NO;

   otherDate = anObject;
   return (([otherDate yearOfCommonEra] == year) &&
           ([otherDate monthOfYear] == month) &&
           ([otherDate dayOfMonth] == day) &&
           ([otherDate minuteOfDay] == minute));
}



- (NSComparisonResult) compare: (id) anObject
{
   Date *otherDate;
   int otherYear, otherMonth, otherDay;
   int otherMinute;

   if (anObject == self) return NSOrderedSame;
   NSParameterAssert ([anObject isKindOfClass: [self class]]);
   otherDate = anObject;

   if (year < (otherYear = [otherDate yearOfCommonEra])) {
      return NSOrderedAscending;
   } else if (year   > otherYear) {
      return NSOrderedDescending;
   } else if (month  < (otherMonth = [otherDate monthOfYear])) {
      return NSOrderedAscending;
   } else if (month  > otherMonth) {
      return NSOrderedDescending;
   } else if (day    < (otherDay = [otherDate dayOfMonth])) {
      return NSOrderedAscending;
   } else if (day    > otherDay) {
      return NSOrderedDescending;
   } else if (minute < (otherMinute = [anObject minuteOfDay])) {
      return NSOrderedAscending;
   } else if (minute > otherMinute) {
      return NSOrderedDescending;
   } else {
      return NSOrderedSame;
   }
}



@end
