/*
  EmailMessage.m
  Copyright (c) 1996-1999 Dwight D. Everhart.  All rights reserved.

  This source code has been released to the public.  Use is governed
  by the MiscKit license, which basically means you can use it for
  whatever purpose you like, commercial or non-commercial.  This code
  is provided AS IS.  No warranty is expressed or implied.
*/

#import "EmailMessage.h"
#import "setInstance.h"
#import <Foundation/NSBundle.h>
#import <Foundation/NSDictionary.h>
#import <Foundation/NSEnumerator.h>
#import <Foundation/NSException.h>
#import <Foundation/NSFileHandle.h>
#import <Foundation/NSString.h>
#import <Foundation/NSTask.h>
#import <Foundation/NSValue.h>


NSString *EmailMessageSendFailureException = @"EmailMessageSendFailureException";

#if defined (unix)
static NSString *emailCommand = @"/usr/lib/sendmail -bm -em -i";
#elif defined (WIN32)
static NSString *emailCommand = @"";
#else
#error Do not know how to send email on this platform.
#endif



@implementation EmailMessage



#ifdef WIN32
+ (void) initialize
{
   NSString *postieExe;

   [super initialize];

   postieExe = [[NSBundle bundleForClass: self] pathForResource: @"postie" ofType: @"exe"];
#warning We should try to find a way to send local email without using an SMTP server.
#warning For now, the SMTP server is hardcoded to mailhost.
   [self setEmailCommand: [postieExe stringByAppendingString: @" -host:mailhost -t"]];
}
#endif



+ (void) setEmailCommand: (NSString *) command
{
   setInstance (emailCommand, command);
}



+ (NSString *) emailCommand
{
   return emailCommand;
}



+ (id) emailMessageWithBody: (NSString *) bodyContents sentTo: (NSString *) to
  withSubject: (NSString *) subject
{
   return [[[self alloc] initWithBody: bodyContents
            sentTo: to withSubject: subject] autorelease];
}



+ (void) sendMessage: (NSString *) bodyContents to: (NSString *) to
  withSubject: (NSString *) subject
{
   [[self emailMessageWithBody: bodyContents sentTo: to withSubject: subject]
      sendMessage];
}



- (id) init
{
   return [self initWithBody: nil sentTo: nil withSubject: nil];
}



- (id) initWithBody: (NSString *) bodyContents sentTo: (NSString *) to
  withSubject: (NSString *) subject
{
   self = [super init];
   if (!self) return nil;

   envelope = [[NSMutableDictionary allocWithZone: [self zone]] initWithCapacity: 8];
   if (!envelope) {
      NSLog (@"EmailMessage: Could not create an envelope dictionary");
      [self dealloc];
      return nil;
   }

   [self setBody: bodyContents];
   [self setTo: to];
   [self setSubject: subject];

   return self;
}



- (void) dealloc
{
   [envelope release];
   [addressee release];
   [body release];
   [super dealloc];
}



- (void) setHeader: (NSString *) header to: (NSString *) contents
{
   NSParameterAssert (header);
   if (contents) {
      [envelope setObject: contents forKey: header];
   } else {
      [envelope removeObjectForKey: header];
   }
}



- (NSString *) contentsOfHeader: (NSString *) header
{
   NSParameterAssert (header);
   return [envelope objectForKey: header];
}



- (void) setFrom: (NSString *) from
{
   [self setHeader: @"From" to: from];
}



- (void) setTo: (NSString *) to
{
   [self setHeader: @"To" to: to];
}



- (void) setCc: (NSString *) cc
{
   [self setHeader: @"Cc" to: cc];
}



- (void) setSubject: (NSString *) subject
{
   [self setHeader: @"Subject" to: subject];
}



- (void) setReplyTo: (NSString *) replyTo
{
   [self setHeader: @"Reply-To" to: replyTo];
}



- (void) setSender: (NSString *) sender
{
   [self setHeader: @"Sender" to: sender];   
}



- (void) setAddressee: (NSString *) anAddressee
{
   setInstance (addressee, anAddressee);
}



- (void) setBody: (NSString *) bodyContents
{
   setInstance (body, bodyContents);
}



- (NSString *) from
{
   return [self contentsOfHeader: @"From"];
}



- (NSString *) to
{
   return [self contentsOfHeader: @"To"];
}



- (NSString *) cc
{
   return [self contentsOfHeader: @"Cc"];
}



- (NSString *) subject
{
   return [self contentsOfHeader: @"Subject"];
}



- (NSString *) replyTo
{
   return [self contentsOfHeader: @"ReplyTo"];
}



- (NSString *) sender
{
   return [self contentsOfHeader: @"Sender"];
}



- (NSString *) addressee
{
   return addressee;
}



- (NSString *) body
{
   return body;
}



- (void) sendMessage
{
   [self sendMessageUsingCommand: emailCommand];
}



- (void) sendMessageUsingCommand: (NSString *) command
{
   NSMutableString *messageHeader;
   NSEnumerator *envelopeEnumerator;
   NSString *header, *headerContents;
   NSArray *commandComponents;
   NSTask *sendEmailTask;
   NSPipe *inputPipe;
   NSFileHandle *writeHandle;
   int status;

   NSParameterAssert (command);
   NSParameterAssert ([self to] || [self addressee]);

   messageHeader = [NSMutableString stringWithCapacity: 1024];
   envelopeEnumerator = [envelope objectEnumerator];
   while ((header = [envelopeEnumerator nextObject])) {
      headerContents = [envelope objectForKey: header];
      [messageHeader appendFormat: @"%@: %@\n", header, headerContents];
   }

   [messageHeader appendString: @"\n"];

   inputPipe = [NSPipe pipe];
   writeHandle = [inputPipe fileHandleForWriting];
   sendEmailTask = [[NSTask allocWithZone: [self zone]] init];
   commandComponents = [command componentsSeparatedByString: @" "];
   [sendEmailTask setLaunchPath: [commandComponents objectAtIndex: 0]];
   [sendEmailTask setStandardInput: inputPipe];
   [sendEmailTask setArguments:
      [commandComponents arrayByAddingObject: addressee ? addressee : [self to]]];

   [sendEmailTask launch];
   [writeHandle writeData:
      [messageHeader dataUsingEncoding: NSNonLossyASCIIStringEncoding]];
   [writeHandle writeData: [body dataUsingEncoding: NSNonLossyASCIIStringEncoding]];
   [writeHandle closeFile];
   [sendEmailTask waitUntilExit];

   status = [sendEmailTask terminationStatus];
   if (status != 0) {
      [[NSException exceptionWithName: EmailMessageSendFailureException
        reason: @"Email command failed"
        userInfo: [NSDictionary dictionaryWithObject: @"ExitStatus"
                   forKey: [NSNumber numberWithInt: status]]] raise];
   }
}



@end
