/*
  Reminder.m
  Copyright (c) 1996-1999 Dwight D. Everhart.  All rights reserved.

  This source code has been released to the public.  Use is governed
  by the MiscKit license, which basically means you can use it for
  whatever purpose you like, commercial or non-commercial.  This code
  is provided AS IS.  No warranty is expressed or implied.
*/

#import "Reminder.h"
#import "Alarm.h"
#import "Appointment.h"
#import "Date.h"
#import <AppKit/AppKit.h>


#define CURRENT_VERSION 2


static BOOL debug = NO;
static BOOL debugVerbosely = NO;


static int firstMinute = 8 * 60;



@implementation Reminder



+ (void)initialize
{
   [super initialize];

   if ([self class] == [Reminder class]) {
      [self setVersion: CURRENT_VERSION];
   }
}



- (id) init
{
   return [self initWithAppointment: nil];
}



- (id) initWithAppointment: (Appointment *) appointment
{
   if (!(self = [super init])) return nil;

   appt = appointment;
   if (debugVerbosely) NSLog (@"Created %@ %p", [[self class] description], self);
   return self;
}



- (id) copyWithZone: (NSZone *) zone
{
   Reminder *clone = NSCopyObject (self, 0, zone);

   if (!clone) return nil;
   /* The new appointment is set in Appointment's copyWithZone: method. */
   clone->appt = nil;
   clone->apptTime = nil;
   clone->alarm = nil;
   return clone;
}



- (void)dealloc
{
   [self stopTimer];
   [NSObject cancelPreviousPerformRequestsWithTarget: self
    selector: @selector (startInitialTimerAndExecuteMissedReminders:) object: nil];
   if (debugVerbosely) NSLog (@"Freed %@ %p", [[self class] description], self);
   [super dealloc];
}



- (Appointment *) appointment
{
   return appt;
}



- (void) setAppointment: (Appointment *) newAppointment
{
   appt = newAppointment;
}



- (remindTime_t) remindTime
{
   return remindTime;
}



- (void) setRemindTime: (remindTime_t) newRemindTime
{
   if (remindTime != newRemindTime) {
      remindTime = newRemindTime;
      [self setCompleted: NO];
   }
}



- (Date *) reminderTime
{
   return [self isCompleted] ? nil : [alarm alarmTime];
}



- (Date *) appointmentTime
{
   return [self isCompleted] ? nil : apptTime;
}



- (BOOL) isEnabled
{
   return rFlags.enabled;
}



- (void)setEnabled:(BOOL)isEnabled
{
   if (rFlags.enabled != isEnabled) {
      rFlags.enabled = isEnabled;
      if (debug) {
         NSLog (@"%@abled %@ %p", isEnabled ? @"En" : @"Dis",
                [[self class] description], self);
      }

      [self setCompleted: !isEnabled];
   }
}



- (BOOL) isCompleted
{
   return !apptTime || !rFlags.enabled;
}



- (void) setCompleted: (BOOL) isCompleted
{
   BOOL wasCompleted = [self isCompleted];

   /*
   If we are not enabled, we must not ever uncomplete ourselves.
    That would produce an inconsistant state.
    */
   if (![self isEnabled]) {
      if (debug && !isCompleted) {
         NSLog (@"Reminder: Changing isCompleted to YES");
      }

      isCompleted = YES;
   }

   /*
      We mustn't do a check to see if the current completion status is
    different from the argument (like the setEnabled: method does),
    because this method is used to reset the apptTime variable,
    (through the setDateToToday message) even when it has already
    been created.
    */
   if (isCompleted) {
      [apptTime release];
      apptTime = nil;
   } else {
      if (!apptTime) {
         apptTime = [[Date allocWithZone:[self zone]] init];
      }

      /* Don't delete this.  If you are tempted, read the above comment. */
      [apptTime setDateToToday];
   }

   if (debug && (isCompleted != wasCompleted)) {
      NSLog (@"%@ompleted %@ %p", isCompleted ? @"C" : @"Unc",
             [[self class] description], self);
   }
}



- (BOOL) getNextReminderTime: (Date *) reminderTime
      includeMissedReminders: (BOOL) includeMissedReminders
   includeImmediateReminders: (BOOL) includeImmediateReminders
{
   int level = [appt scheduleLevel];
   Date *apptEndTime = nil; /* Must initialize. */
   Date *now = [Date now];
   BOOL result = YES;

   while (1) {
      if (![appt getNextOccuranceDate: apptTime]) {
         if (debug) {
            NSLog (@"No more occurances for reminder %@ %p", [[self class] description], self);
         }

         result = NO;
         break;
      }

      [reminderTime setDate: apptTime];
      if (level != ASL_EXACT) [reminderTime setMinute: firstMinute];
      [reminderTime changeMinuteBy: -remindTime];
      if ([reminderTime compare: now] >= (includeImmediateReminders ? 0 : 1)) {
         break;
      }

      if (includeMissedReminders) {
         if (!apptEndTime) {
            apptEndTime = [apptTime copyWithZone:[self zone]];
         } else {
            [apptEndTime setDate: apptTime];
         }

         switch (level) {
            case ASL_NONE:
               break;
            case ASL_YEAR:
               [apptEndTime incrementYear];
               break;
            case ASL_MONTH:
               [apptEndTime incrementMonth];
               break;
            case ASL_WEEK:
               [apptEndTime incrementWeek];
               break;
            case ASL_DAY:
               [apptEndTime incrementDay];
               break;
            case ASL_EXACT:
               [apptEndTime changeMinuteBy: [appt duration]];
               break;
         }

         if ([apptEndTime compare: now] >= 0) {
            if (debug) {
               NSLog (@"Scheduling execution of missed reminder %@ %p",
                      [[self class] description], self);
               NSLog (@"Appointment will occur in %d minutes",
                     [now minutesUntil: apptTime]);
            }
            break;
         }
      }

      if (debug) {
         NSLog (@"Missed reminder %@ %p", [[self class] description], self);
         NSLog (@"Appointment occurred %d minutes ago", [apptTime minutesUntil: now]);
      }

      /* Try to schedule for the next occurance. */
      [apptTime incrementDay];
   }

   [apptEndTime release];
   return result;
}



- (void) startTimer
{
    [self startTimerAndExecuteMissedReminders: NO andImmediateReminders: YES];
}



- (void) startTimerAndExecuteMissedReminders: (BOOL) executeMissedReminders
                       andImmediateReminders: (BOOL) executeImmediateReminders
{
   Date *oldApptTime;
   Date *reminderTime;

   [self stopTimer];
   if (![self isEnabled] || [self isCompleted]) return;

   oldApptTime = [apptTime copyWithZone:[self zone]];
   reminderTime = [apptTime copyWithZone:[self zone]];

   if (![self getNextReminderTime: reminderTime
             includeMissedReminders: executeMissedReminders
          includeImmediateReminders: executeImmediateReminders]) {
      [self setCompleted: YES];
      [oldApptTime release];
      [reminderTime release];
      return;
   }

   [oldApptTime release];

   if (!alarm) {
      alarm = [[Alarm allocWithZone:[self zone]] init];
   } else {
      NSLog (@"Reminder: Alarm was already created");
   }

   [alarm setAlarmTime: reminderTime target: self
            andMessage: @selector (executeReminderAndReschedule:)];
   if (debug) NSLog (@"Created alarm for %@ %p", [[self class] description], self);

   [reminderTime release];
}



- (void) startInitialTimerAndExecuteMissedReminders: (int) executeMissedReminders
{
   [self startTimerAndExecuteMissedReminders: executeMissedReminders
                                andImmediateReminders: executeMissedReminders];
}



- (void) stopTimer
{
   if (alarm) {
      [alarm release];
      alarm = nil;
      if (debug) NSLog (@"Freed alarm for %@ %p", [[self class] description], self);
   }
}



- (void) executeReminderAndReschedule: (id) sender
{
   [self stopTimer];
   if (![self isEnabled] || [self isCompleted]) return;

   if (debug) NSLog (@"Executing %@ %p", [[self class] description], self);

   /* Do the reminder. */
   [self doReminder];

   /*
      Advance the apptTime since we've already issued our reminder
    for this day.  If we don't do this, and the system clock is running
    a little fast, we may issue two reminders instead of one.
    */
   [apptTime incrementDay];

   /* Start the timer for the next occurance. */
   [self startTimer];
}



- (void) doReminder
{
   /* Override this in subclasses. */
}



- (void) objectDidChange: (id) object
{
   if ([self isEnabled]) {
      [self setCompleted: NO];
      [self startTimerAndExecuteMissedReminders: NO andImmediateReminders: NO];
   }
}



- (void) objectWillDealloc: (id) object
{
   appt = nil;
   [self stopTimer];
   [self setCompleted: YES];
}



@end
